from . import cpp
from . import python as pyschema
from . import json_example as jsex
import stringcase
from . import schemawrappers
import yaml
from copy import copy

class TreeWalkerException(Exception):
    pass

class SimpleResolver(cpp.ResolverBaseClass, pyschema.ResolverBaseClass, jsex.SchemaResolverBaseClass):

    def __init__(self, uri, root=None):
        self.uri = uri
        self.root = root
        self.usings = []
        self.rootNs = []

    def cpp_set_root_namespace(self, ns: list):
        self.rootNs = ns

    def cpp_add_using(self, using: list):
        self.usings.append(using)

    def cpp_get_root_namespace(self) -> list:
        return self.rootNs

    def cpp_get_usings(self) -> list:
        return self.usings

    def _get_reference_parts(self, reference) -> tuple:
        pkg = None
        fn = reference.split('#')[0] or None
        path = reference.split('#')[1]
        parts = path.split('/')
        if fn:
            pkg = fn.split('.')[0]
        return (pkg, parts[-2], stringcase.pascalcase(parts[-1]))

    def _get_reference_parts(self, reference) -> dict:
        assert('#' in reference), "Reference '{}' seemed malformed".format(reference)
        url, path = reference.split('#')
        theType, name = path.split('/')[-2:]
        return {
            "url": url,
            "uri": url,
            "path": path,
            "type": theType,
            "PascalType": stringcase.pascalcase(theType),
            "name": name,
            "PascalName": stringcase.pascalcase(name),
            "snake_name": stringcase.snakecase(name),
            "pkg": url is not None and url.split(".")[0] or None,
        }

    def cpp_get_header_dir(self, uri=None) -> str:
        uri = uri or self.uri
        pkg = uri.split(".")[0]
        return pkg

    def cpp_get_filename_base(self, reference) -> str:
        parts = self._get_reference_parts(reference)
        if parts['pkg'] not in [None, self.uri]:
            header_dir = self.cpp_get_header_dir(parts['uri'])
            fn_base = "{0}/{1}_{2}".format(header_dir, parts['type'].rstrip('s'), parts['snake_name'])
        else:
            fn_base = "{0}_{1}".format(parts['type'].rstrip('s'), parts['snake_name'])
        return fn_base

    def cpp_get_header(self, reference) -> str:
        header = "{}.hpp".format(self.cpp_get_filename_base(reference))
        return header

    def cpp_get_namespace(self, reference) -> list:
        parts = self._get_reference_parts(reference)
        ns = copy(self.rootNs)
        if parts['pkg'] is not None:
            ns.append(stringcase.lowercase(parts['pkg']))
        else:
            ns.append(self.uri)
        ns.append(stringcase.lowercase(parts['type'].rstrip('s')))
        return ns

    def cpp_get_name(self, reference) -> str:
        parts = self._get_reference_parts(reference)
        return parts['PascalName']

    def py_include_statement(self, reference) -> str:
        ref = self._get_reference_parts(reference)
        if ref['pkg'] is not None:
            return "import {pkg}.{type}_{name}".format(**ref)
        else:
            return "import {type}_{name}".format(**ref)

    def py_class_name(self, reference) -> str:
        ref = self._get_reference_parts(reference)
        if ref['pkg'] is not None:
            return "{pkg}.{type}_{name}.{PascalName}".format(**ref)
        else:
            return "{type}_{name}.{PascalName}".format(**ref)

    def py_filename(self, reference) -> str:
        ref = self._get_reference_parts(reference)
        return "{type}_{name}.py".format(**ref)

    def _walk_through_tree(self, tree, path) -> dict:
        assert(tree is not None), "No tree to walk through"
        walker = tree
        for p in [p for p in path.split('/') if len(p) > 0]:
            if p not in walker:
                treeId = 'id' in tree and tree['id'] or 'UNKNOWN'
                raise TreeWalkerException("Could not resolve {} from {}".format(path, treeId))
            walker = walker[p]
        return walker

    def get_document(self, uri, encoding=None) -> dict:
        if '#' in uri:
            uri = uri.split('#')[0]
        with open(uri, 'r') as fp:
            if encoding == 'json' or (encoding is None and 'json' in uri):
                import json
                return json.load(fp)
            else:
                return yaml.load(fp, Loader=yaml.FullLoader)

    def _get_root(self, reference, root):
        parts = self._get_reference_parts(reference)
        if parts['uri']:
            root_doc = self.get_document(parts['uri'])
        elif root is not None:
            root_doc = root
        else:
            root_doc = self.root
        return root_doc

    def get_json(self, reference, root=None) -> dict:
        parts = self._get_reference_parts(reference)
        root_doc = self._get_root(reference, root)
        try:
            json_doc = self._walk_through_tree(root_doc, parts['path'])
        except TreeWalkerException as e:
            print(f"{self} Error in resolving {reference}: {e}")
            raise
        else:
            return json_doc

    def get_schema(self, reference, root=None) -> schemawrappers.SchemaBase:
        schema = schemawrappers.SchemaFactory(self.get_json(reference, root=root))
        schema.root = self._get_root(reference, root)
        return schema




import json
import abc
import stringcase
import os.path

from . import templator
from . import schemawrappers
from . import json_example

class ResolverBaseClass(abc.ABC):

    @abc.abstractmethod
    def py_include_statement(self, reference):
        """Should return the include statement needed to acquire the object representing the
        schema pointed to at `reference`.  Example, "from schema_foo import Foo"
        """
        pass

    @abc.abstractmethod
    def py_class_name(self, reference):
        """Should return the class name for the object representing the schema pointed to at `reference`.
        For example, "schema_foo.Foo"
        """
        pass

    @abc.abstractmethod
    def py_filename(self, reference):
        """Should return the name of the filename holding the python class representing the schema pointed to
        at `reference`.  For example, "schema_foo.py"
        """
        pass


class GeneratorFromSchema(object):

    def __init__(self, output_dir, resolver=None):
        self.output_dir = output_dir
        self.resolver = resolver

    def GetDeps(self, schema):
        return []

    def Generate(self, schema, class_name, filename_base):
        generator = templator.Generator('jsonschemacodegen.templates.python', self.output_dir)
        args = {
            "Name": class_name,
            "schema": schemawrappers.SchemaFactory(schema),
        }
        return generator.RenderTemplate("file.py.jinja2", output_name="{}.py".format(filename_base), resolver=self.resolver, **args)

    def Examples(self, schema, root):
        wrapped_schema = schemawrappers.SchemaFactory(schema, root)
        number_of_examples = wrapped_schema.GetExampleCombos(self.resolver)
        examples = []
        examples.append(wrapped_schema.Example(self.resolver, schemawrappers.ExampleIndex(0)))
        examples.append(wrapped_schema.Example(self.resolver, schemawrappers.ExampleIndex(-1)))
        show_examples = min(number_of_examples, 10)
        example_step = int(number_of_examples/show_examples)
        index = example_step
        for _ in range(0, show_examples):
            examples.append(wrapped_schema.Example(self.resolver, schemawrappers.ExampleIndex(index)))
            index += example_step
        return sorted(list(set([json.dumps(x) for x in examples])))

    def GenerateTest(self, schema, root, class_name, filename_base, path):
        filename = self.resolver.py_test_filename(path)
        generator = templator.Generator('jsonschemacodegen.templates.python', os.path.join(self.output_dir, os.path.dirname(filename)))
        wrapped_schema = schemawrappers.SchemaFactory(schema, root)
        args = {
            "Name": class_name.split('.')[-1],
            "schema": wrapped_schema,
            "examples": self.Examples(schema, root),
            "class": class_name,
            "path": path,
            "objType": path.split("/")[-2]
        }
        return generator.RenderTemplate("test.py.jinja2", output_name=os.path.basename(filename), resolver=self.resolver, **args)

    def GenerateFromPath(self, schema, path):
        assert(self.resolver)
        class_name = self.resolver.py_class_name(path).split('.')[-1]
        filename_base = self.resolver.py_filename(path)
        return self.Generate(schema, class_name, filename_base)

    def GenerateTestFromPath(self, schema, root, path):
        assert(self.resolver)
        class_name = self.resolver.py_class_name(path)
        filename_base = self.resolver.py_filename(path)
        return self.GenerateTest(schema, root, class_name, filename_base, path)


import collections
import random

def bitsNeededForNumber(num):
    bits = 0
    while num > 0:
        num = num >> 1
        bits += 1
    return bits

def rightBitMask(bits):
    mask = 1
    mask = mask << bits
    mask -= 1
    return mask

class ExampleIndex(object):

    def __init__(self, index):
        self.index = index
        self.current = index

    def _full(self):
        return self.index == -1

    def BooleanChoice(self):
        if self._full():
            return True
        choice = self.current & 0x01
        self.current = self.current >> 1
        return bool(choice)

    def Choice(self, population):
        return population[self.Number(len(population))-1]

    def Number(self, maximum):
        if self._full() or maximum == 0:
            return maximum
        bits = bitsNeededForNumber(maximum)
        mask = rightBitMask(bits)
        choice = self.current & mask
        self.current = self.current >> bits
        return choice % maximum


class SchemaBase(collections.UserDict):

    def __init__(self, initialdata, root=None):
        super().__init__(initialdata)
        self.root = root

    def CppIncludes(self, resolver):
        return {
            '"rapidjson/document.h"',
            "<exception>"
        }

    def Resolve(self, resolver):
        return self

    def IsReadOnly(self):
        return ('readOnly' in self.data) and (self.data['readOnly'] is True)
    
    def IsWriteOnly(self):
        return ('writeOnly' in self.data) and (self.data['writeOnly'] is True)

    def GetExampleCombos(self, resolver) -> int:
        if 'examples' in self.data:
            return len(self.data['examples'])
        return 1
    
    def Example(self, resolver, index: ExampleIndex):
        if 'example' in self.data:
            return self.data['example']
        if 'examples' in self.data:
            return index.Choice(self.data['examples'])
        if 'default' in self.data:
            return self.data['default']
        return self.AnExample(resolver, index)

class Reference(SchemaBase):

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        if resolver:
            incs.update({'"'+resolver.cpp_get_header(self.data['$ref'])+'"'})
        return incs

    def Resolve(self, resolver):
        return resolver.get_schema(self.data['$ref'], self.root)

    def GetExampleCombos(self, resolver) -> int:
        return self.Resolve(resolver).GetExampleCombos(resolver)

    def Example(self, resolver, index: ExampleIndex):
        return self.Resolve(resolver).Example(resolver, index)


class ObjectSchema(SchemaBase):

    def __init__(self, initialdata, root=None):
        super().__init__(initialdata, root)
        if 'properties' in self.data:
            assert(isinstance(self.data['properties'], dict))
        else:
            self.data['properties'] = dict()

    def GetPropertySchemas(self):
        props = {}
        for n, p in self.data['properties'].items():
            props[n] = SchemaFactory(p, self.root)
        return props

    def PropertyKeys(self):
        return [a for a in self.data['properties'].keys()]

    def PropertyValues(self):
        return [a for a in self.data['properties'].values()]

    def __getitem__(self, key):
        if key == 'properties':
            return self.GetPropertySchemas()
        else:
            return self.data[key]

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        incs.update({"<boost/optional.hpp>"})
        for _, ps in self.GetPropertySchemas().items():
            incs.update(ps.CppIncludes(resolver))
        return incs

    # TODO: Need something that specifies that this is 'required' for init
    def RequiredList(self, default_negates_required=True):
        theList = []
        for propName, propSchema in self.data['properties'].items():
            if ('required' in self.data and propName in self.data['required']) and not (default_negates_required and 'default' in propSchema):
                theList.append((propName, SchemaFactory(propSchema, self.root)))
        return theList

    # TODO: Need something that specifies that this is 'required' for init
    def UnRequiredList(self, default_negates_required=True):
        theList = []
        for propName, propSchema in self.data['properties'].items():
            if default_negates_required and 'default' in propSchema:
                theList.append((propName, SchemaFactory(propSchema, self.root)))
            elif 'required' not in self.data:
                theList.append((propName, SchemaFactory(propSchema, self.root)))
            elif propName not in self.data['required']:
                theList.append((propName, SchemaFactory(propSchema, self.root)))
        return theList

    def GetExampleCombos(self, resolver) -> int:
        combos = 1
        for _, item in self.RequiredList(default_negates_required=False):
            combos *= item.GetExampleCombos(resolver)
        for _, item in self.UnRequiredList(default_negates_required=False):
            combos *= 2
            combos *= item.GetExampleCombos(resolver)
        return combos

    def AnExample(self, resolver, index: ExampleIndex):
        ret = {}
        for name, item in self.RequiredList(default_negates_required=False):
            ret[name] = item.Example(resolver, index)
        for name, item in self.UnRequiredList(default_negates_required=False):
            if index.BooleanChoice():
                ret[name] = item.Example(resolver, index)
        return ret

class StringSchema(SchemaBase):

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        incs.update({"<string>", "<boost/functional/hash.hpp>"})
        if 'pattern' in self.data:
            incs.add("<regex>")
        if 'format' in self.data:
            if self.data['format'] == 'uuid':
                incs.update({"<boost/uuid/uuid.hpp>", "<boost/uuid/random_generator.hpp>", "<boost/uuid/uuid_io.hpp>"})
            elif self.data['format'] == 'date-time':
                incs.update({"<boost/optional.hpp>", "<boost/date_time/posix_time/posix_time.hpp>", "<boost/algorithm/string/replace.hpp>"})
        return incs

    def AnExample(self, resolver, index: ExampleIndex):
        if 'format' in self.data:
            if self.data['format'] == 'uuid':
                return "a18dfb2c-2b17-4a19-85e0-5c6c8d6a89e8"
            elif self.data['format'] == 'date-time':
                return "2002-10-02T15:00:00Z"
        maxLen = 'maxLength' in self.data and self.data['maxLength'] or 6
        minLen = 'minLength' in self.data and self.data['minLength'] or min(maxLen, 6)
        theString = ("string"*minLen)[:minLen]
        return theString


class StringEnumSchema(StringSchema):
    
    def GetExampleCombos(self, resolver) -> int:
        if 'example' in self.data or 'examples' in self.data or 'default' in self.data:
            return super().GetExampleCombos(resolver)
        return len(self.data['enum'])
    
    def AnExample(self, resolver, index: ExampleIndex):
        return index.Choice(self.data['enum'])


class NumberSchema(SchemaBase):

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        incs.update({"<boost/lexical_cast.hpp>", "<boost/functional/hash.hpp>"})
        return incs

    def AnExample(self, resolver, index: ExampleIndex):
        for k in ['minimum', 'maximum']:
            if k in self.data:
                return self.data[k]
        if 'exclusiveMinimum' in self.data:
            return self.data['exclusiveMinimum'] + (self.data['type'] == 'integer' and 1 or 0.000001)
        if 'exclusiveMaximum' in self.data:
            return self.data['exclusiveMaximum'] - (self.data['type'] == 'integer' and 1 or 0.000001)
        return 1

class BooleanSchema(SchemaBase):

    def GetExampleCombos(self, resolver) -> int:
        if 'example' in self.data or 'examples' in self.data or 'default' in self.data:
            return super().GetExampleCombos(resolver)
        return 2

    def AnExample(self, resolver, index: ExampleIndex):
        return index.Choice([True, False])


class NullSchema(SchemaBase):

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        incs.update({"<boost/none.hpp>"})
        return incs

    def AnExample(self, resolver, index: ExampleIndex):
        return None

class ArraySchema(SchemaBase):
    
    def GetItemSchema(self):
        return SchemaFactory(self.data['items'], self.root)

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        incs.update({"<vector>"})
        incs.update(self.GetItemSchema().CppIncludes(resolver))
        return incs

    def GetExampleCombos(self, resolver) -> int:
        if 'example' in self.data or 'examples' in self.data or 'default' in self.data:
            return super().GetExampleCombos(resolver)
        left = 'minItems' in self.data and int(self.data['minItems']) or 0
        right = 'maxItems' in self.data and int(self.data['maxItems']) or (left + 3)
        combos = bitsNeededForNumber((right - left)+1)
        combos *= self.GetItemSchema().GetExampleCombos(resolver)
        return combos

    def AnExample(self, resolver, index: ExampleIndex):
        ret = []
        left = 'minItems' in self.data and int(self.data['minItems']) or 0
        right = 'maxItems' in self.data and int(self.data['maxItems']) or (left + 3)
        for _ in range(0, left+index.Number(right-left)):
            ret.append(self.GetItemSchema().Example(resolver, index))
        return ret

class CombinatorSchemaBase(SchemaBase):
    
    def __init__(self, name, initialdata, root=None):
        super().__init__(initialdata, root)
        self.name = name

    def GetComponents(self):
        return [SchemaFactory(s, self.root) for s in self.data[self.name]]

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        for c in self.GetComponents():
            incs.update(c.CppIncludes(resolver))
        return incs

    def __getitem__(self, key):
        if key.upper() in str(type(self)).upper():
            return self.GetComponents()
        return self.data[key]


class OneOfSchema(CombinatorSchemaBase):
    
    def __init__(self, initialdata, root=None):
        super().__init__('oneOf', initialdata, root)
        assert(isinstance(self.data['oneOf'], list))

    def CppIncludes(self, resolver=None):
        incs = super().CppIncludes(resolver=resolver)
        incs.update({"<boost/variant.hpp>", "<boost/functional/hash.hpp>"})
        return incs

    def GetCommonType(self, resolver):
        commonType = None
        for comp in self.GetComponents():
            subSchema = comp.Resolve(resolver)
            if 'type' in subSchema:
                if commonType is None:
                    commonType = subSchema['type']
                elif commonType != subSchema['type']:
                    return None
        return commonType

    def GetExampleCombos(self, resolver) -> int:
        if 'example' in self.data or 'examples' in self.data or 'default' in self.data:
            return super().GetExampleCombos(resolver)
        return bitsNeededForNumber(len(self.GetComponents()) - 1)

    def AnExample(self, resolver, index: ExampleIndex):
        return index.Choice(self.GetComponents()).Example(resolver, index)


class AllOfSchema(CombinatorSchemaBase):

    def __init__(self, initialdata, root=None):
        super().__init__('allOf', initialdata, root)
        assert(isinstance(self.data['allOf'], list))

    def AnExample(self, resolver, index: ExampleIndex):
        ret = {}
        for comp in self.GetComponents():
            ret.update(comp.Example(resolver, index))
        return ret


class AnyOfSchema(CombinatorSchemaBase):

    def __init__(self, initialdata, root=None):
        super().__init__('anyOf', initialdata, root)
        assert(isinstance(self.data['anyOf'], list))

    def GetExampleCombos(self, resolver) -> int:
        if 'example' in self.data or 'examples' in self.data or 'default' in self.data:
            return super().GetExampleCombos(resolver)
        numberComponents = len(self.GetComponents())
        combos = bitsNeededForNumber(numberComponents - 1)
        combos << (numberComponents - 1)
        return combos
    
    def AnExample(self, resolver, index: ExampleIndex):
        ret = {}
        gotOne = False
        for comp in self.GetComponents():
            if index.BooleanChoice():
                gotOne = True
                ret.update(comp.Example(resolver, index))
        if not gotOne:
            comp = index.Choice(self.GetComponents())
            ret.update(comp.Example(resolver, index))
        return ret

def SchemaFactory(schema, root=None):
    if 'type' in schema:
        if schema['type'] == 'string':
            if 'enum' in schema:
                return StringEnumSchema(schema, root)
            return StringSchema(schema, root)
        elif schema['type'] == 'number' or schema['type'] == 'integer':
            return NumberSchema(schema, root)
        elif schema['type'] == 'boolean':
            return BooleanSchema(schema, root)
        elif schema['type'] == 'null':
            return NullSchema(schema, root)
        elif schema['type'] == 'object':
            return ObjectSchema(schema, root)
        elif schema['type'] == 'array':
            return ArraySchema(schema, root)
        else:
            raise NotImplementedError(f"The type '{schema['type']}' is not implemented")
    elif 'allOf' in schema:
        return AllOfSchema(schema, root)
    elif 'anyOf' in schema:
        return AnyOfSchema(schema, root)
    elif 'oneOf' in schema:
        return OneOfSchema(schema, root)
    elif '$ref' in schema:
        return Reference(schema, root)
    else:
        raise NotImplementedError(str(schema))



import json
import random
import abc
from copy import copy
from . import schemawrappers

class SchemaResolverBaseClass(abc.ABC):

    @abc.abstractmethod
    def get_schema(self, reference, root=None):
        """Given a reference, returns a wrapped schema object.
        """
        pass

    @abc.abstractmethod
    def get_json(self, reference, root=None) -> dict:
        pass

    @abc.abstractmethod
    def get_document(self, reference):
        pass


class GeneratorFromSchema(object):

    def __init__(self, resolver=None):
        self.resolver = resolver
        self.includeMode = 'all'
        self.seed = 0

    def GetExampleOr(self, schema, default):
        if 'example' in schema:
            return schema['example']
        elif 'examples' in schema:
            exampleList = sorted(schema['examples'])
            return exampleList[0]
        elif 'enum' in schema:
            enums = sorted(schema['enum'])
            return enums[0]
        else:
            return default

    def GetNumber(self, schema):
        return self.GetExampleOr(schema, schema['type'] == 'integer' and 1 or 3.14)

    def GetObject(self, root, schema, base=None):
        base = base or {}
        for propName, propSchema in schema['properties'].items():
            if (self.includeMode == 'required') and (('required' not in schema) or (propName not in schema['required'])):
                continue
            thing = self.GetThing(root, propSchema)
            base[propName] = thing
        default = None
        if 'default' in schema:
            default = schema['default']
        elif 'defaults' in schema and isinstance(schema['defaults'], list):
            default = schema['defaults'][0]
        if default is not None:
            base.update(default)
        return base

    def GetArray(self, root, schema):
        base = []
        defaultMin = self.includeMode == 'all' and 1 or 0
        minItems = 'minItems' in schema and schema['minItems'] or defaultMin
        for _ in range(0, minItems):
            base.append(self.GetThing(root, schema['items']))
        return base

    def GetThing(self, root, schema, base=None):
        root_doc = root
        base = base or {}
        if '$ref' in schema:
            if len(schema['$ref'].split('#')[0]) > 0:
                root_doc = self.resolver.get_document(schema['$ref'])
            schema = self.resolver.get_schema(reference=schema['$ref'], root=root)
        if 'allOf' in schema:
            obj = base
            for opt in schema['allOf']:
                obj = self.GetThing(root_doc, opt, base=obj)
            return obj
        elif 'anyOf' in schema:
            obj = base
            if self.includeMode == 'required':
                return obj
            for opt in schema['anyOf']:
                obj = self.GetThing(root_doc, opt, base=obj)
            return obj
        elif 'oneOf' in schema:
            random.seed(self.seed)
            thing = self.GetThing(root_doc, random.choice(schema['oneOf']))
            return thing
        elif 'type' not in schema:
            raise NotImplementedError(schema)
        elif schema['type'] in ['integer', 'number']:
            return self.GetNumber(schema)
        elif schema['type'] == 'string':
            return self.GetExampleOr(schema, 'string')
        elif schema['type'] == 'null':
            return None
        elif schema['type'] == 'boolean':
            return self.GetExampleOr(schema, True)
        elif schema['type'] == 'object':
            return self.GetObject(root_doc, schema, base)
        elif schema['type'] == 'array':
            return self.GetArray(root_doc, schema)
        else:
            raise NotImplementedError

    def GenerateSome(self, root, schema, run, includeMode) -> set:
        self.includeMode = (includeMode in ['all', 'required']) and includeMode or 'all'
        examples = set()
        for i in range(0, run):
            self.seed = i
            thing = self.GetThing(root, schema)
            examples.add(json.dumps(thing, sort_keys=True))
        return examples

    def GenerateFull(self, root, schema, run=100) -> set:
        assert(not isinstance(schema, schemawrappers.Reference))
        if isinstance(schema, schemawrappers.SchemaBase):
            schemaJsonText = json.dumps(schema.data)
        else:
            schemaJsonText = json.dumps(schema)
        run = max(2, schemaJsonText.count('oneOf')) * 10
        return self.GenerateSome(root, schema, run, 'all')

    def GenerateLimited(self, root, schema, run=2) -> set:
        if isinstance(schema, schemawrappers.SchemaBase):
            schemaJsonText = json.dumps(schema.data)
        else:
            schemaJsonText = json.dumps(schema)
        run = max(1, schemaJsonText.count('oneOf')) * 5
        return self.GenerateSome(root, schema, run, 'required')
    
    def Generate(self, root, schema) -> list:
        full = self.GenerateFull(root, schema)
        full.update(self.GenerateLimited(root, schema))
        return [json.loads(s) for s in full]




import os
import jinja2
import stringcase
import re

PLANTUML_REPLACE = re.compile(r"```plantuml\n(.*)```", re.DOTALL|re.MULTILINE)

class Generator(object):

    def __init__(self, templatePkg, outputDir):
        self.templatePkg = templatePkg
        self.outputDir = outputDir
        self.generatedFiles = []
        self.jinjaEnvironment = None

    def GetJinjaEnvironment(self):
        def IsOfType(obj, theType):
            return theType in str(type(obj))

        def AddLeadingUnderscore(s: str):
            return "_%s" % (s)

        def Privatize(s: str):
            return AddLeadingUnderscore(stringcase.camelcase(s))

        def QuoteIfString(s: str, condition):
            if condition == 'string' or isinstance(condition, str):
                return '"%s"' % (s)
            return s

        def Enumify(s: str):
            if s[0].isnumeric():
                return "_"+stringcase.constcase(s)
            return stringcase.constcase(s)

        def UpperCamelCase(s: str):
            return stringcase.pascalcase(stringcase.snakecase(s)).replace('_', '')

        def MdIndent(s: str, width: int):
            indention = " " * width
            newline = "\n"

            s += newline  # this quirk is necessary for splitlines method

            lines = s.splitlines()
            rv = lines.pop(0)

            if lines:
                rv += newline + newline.join(
                    indention + line if (line and not line.strip().startswith('<')) else line for line in lines
                )

            return rv

        def Italics(s: str):
            if s and s is not None and s != 'None' and len(s) > 0:
                return "_%s_" % (s)
            else:
                return ''

        def Doxygenify(s: str):
            if "```plantuml" in s:
                return PLANTUML_REPLACE.sub(r"\\startuml\n\1\\enduml", s)
            return s

        if self.jinjaEnvironment is None:
            env = jinja2.Environment(loader=jinja2.PackageLoader(self.templatePkg, ''),
                                     extensions=['jinja2.ext.do'])
            env.filters['UpperCamelCase'] = UpperCamelCase
            env.filters['PascalCase'] = lambda s: stringcase.pascalcase(stringcase.snakecase(s))
            env.filters['CONST_CASE'] = lambda s : stringcase.constcase(str(s))
            env.filters['snake_case'] = stringcase.snakecase
            env.filters['camelCase'] = stringcase.camelcase
            env.filters['mdindent'] = MdIndent
            env.filters['type'] = type # For debugging
            env.filters['doxygen'] = Doxygenify
            env.filters['italics'] = Italics
            env.filters['quotestring'] = QuoteIfString
            env.filters['dir'] = dir # For debug
            env.filters['privatize'] = Privatize
            env.filters['enumify'] = Enumify
            env.tests['oftype'] = IsOfType
            self.jinjaEnvironment = env
        return self.jinjaEnvironment

    def GetRenderedTemplate(self, template_name, **kwargs):
        template = self.GetJinjaEnvironment().get_template(template_name)
        rendered = template.render(kwargs)
        return rendered

    def RenderTemplate(self, template_name, output_name = None, **kwargs):
        output_filename = output_name or ".".join(template_name.split(".")[:-1])
        output_file = os.path.join(self.outputDir, output_filename)
        rendered = self.GetRenderedTemplate(template_name, **kwargs)
        with open(output_file, "w") as fp:
            fp.write(rendered)
        self.generatedFiles.append(output_filename)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fp:
    requirements = fp.read().split('\n')

import jsonschemacodegen._version as _version

setup(name='json-schema-codegen',
      version=_version.__version__,
      url='http://github.com/pearmaster/json-schema-codegen',
      author='Jacob Brunson',
      author_email='pypi@jacobbrunson.com',
      description="Generate C++ or Python structures from JSON-Schema",
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='GPLv2',
      packages=[
          'jsonschemacodegen',
          'jsonschemacodegen.templates.cpp',
          'jsonschemacodegen.templates.python',
          'jsonschemacodegen.templates.markdown',
      ],
      package_data={
            'jsonschemacodegen.templates.cpp': ['*.jinja2'],
            'jsonschemacodegen.templates.python': ['*.jinja2'],
            'jsonschemacodegen.templates.markdown': ['*.jinja2'],
      },
      zip_safe=False,
      install_requires=requirements,
      include_package_data=True,
      python_requires='>=3.7',
)


import json
import abc
import stringcase
import os

from . import templator
from . import schemawrappers

class ResolverBaseClass(abc.ABC):

    def cpp_resolve_namespace(self, ns: list, append='::') -> str:
        """ Figures out what a namespace should be for C++.
        @param ns is the full namespace of the object that should be resolved.  For example, if the object is:
            d::e::f::Goo, then the ns param should be ['d', 'e', 'f'] with 'Goo' being the object name.
        @param append is a string that will be appended to the result only if there is some part of a namespace left.
        @returns a string that should go before the object.  Using the described parameters, the result
            would be "f::"
        """
        assert(isinstance(ns, list))
        usings = self.cpp_get_usings()
        for n in ns:
            assert(isinstance(n, str)), "namespace is %s" % (ns)
        ret = ns
        nsString = "||".join(ns)
        for using in usings:
            if nsString.startswith("||".join(using)):
                abbr = ns[len(using):]
                if len(abbr) < len(ret):
                    ret = abbr
        if len(ret) > 0:
            return "::".join(ret)+append
        else:
            return ""

    def append_to_namespace(self, ns: list, appendage) -> list:
        assert(isinstance(ns, list)), "Namespace isn't list it is %s" % (ns)
        newNs = ns.copy()
        newNs.append(appendage)
        return newNs

    @abc.abstractmethod
    def cpp_get_filename_base(self, reference):
        pass

    @abc.abstractmethod
    def cpp_get_header(self, reference: str) -> str:
        """ Given a $ref reference from a schema, return the name/path of the header file that should be included for the declaration
        of the represented object.
        """
        pass

    @abc.abstractmethod
    def cpp_get_namespace(self, reference: str) -> list:
        """Given a reference and the current usings statements, return the namespace of the object pointed to by the reference.
        If the namespace is not empty, also append the append string
        """
        pass

    @abc.abstractmethod
    def cpp_get_name(self, reference: str) -> str:
        """Given a reference, return the C++ object name pointed to by the reference.
        """
        pass

    def cpp_get_ns_name(self, reference: str) -> str:
        ns = self.cpp_get_namespace(reference)
        name = self.cpp_get_name(reference)
        if ns is None:
            return 
        return "{}::{}".format("::".join(ns), name)

    @abc.abstractmethod
    def cpp_get_root_namespace(self) -> list:
        pass

    @abc.abstractmethod
    def cpp_get_usings(self) -> list:
        pass


class GeneratorFromSchema(object):

    def __init__(self, src_output_dir, header_output_dir, resolver):
        self.output_dir = {
            "src": src_output_dir,
            "header": header_output_dir,
        }
        assert(isinstance(resolver, ResolverBaseClass)), "Resolver is %s" % (resolver)
        self.resolver = resolver

    def _make_sure_directory_exists(self, output_key, dir_path):
        d = os.path.join(self.output_dir[output_key], dir_path)
        if not os.path.exists(d):
            os.makedirs(d)

    def GetDeps(self, schema):
        return schema.CppIncludes(self.resolver)

    def Generate(self, schema, path):
        retval = [None, None]
        srcGenerator = templator.Generator('jsonschemacodegen.templates.cpp', self.output_dir['src'])
        headerGenerator = templator.Generator('jsonschemacodegen.templates.cpp', self.output_dir['header'])
        args = {
            "Name": self.resolver.cpp_get_name(path),
            "schema": schemawrappers.SchemaFactory(schema),
        }
        headerFilename = self.resolver.cpp_get_header(path)
        self._make_sure_directory_exists('header', os.path.dirname(headerFilename))
        if '$ref' not in schema:
            srcFileName = "{}.cpp".format(self.resolver.cpp_get_filename_base(path))
            self._make_sure_directory_exists('src', os.path.dirname(srcFileName))
            srcGenerator.RenderTemplate("source.cpp.jinja2", 
                srcFileName, 
                deps=['"{}"'.format(self.resolver.cpp_get_header(path))], 
                usings=self.resolver.cpp_get_usings(),
                ns=self.resolver.cpp_get_namespace(path),
                resolver=self.resolver,
                **args)
            retval[0] = srcFileName
        headerGenerator.RenderTemplate("header.hpp.jinja2", 
            headerFilename, 
            ns=self.resolver.cpp_get_namespace(path),
            deps=self.GetDeps(args['schema']), 
            resolver=self.resolver, 
            **args)
        retval[1] = headerFilename
        return tuple(retval)

